@@ -56,7 +56,7 @@ void pl_shader_free(struct pl_shader **psh)
     TA_FREEP(psh);
 }
 
-void pl_shader_reset(struct pl_shader *sh, uint8_t ident, uint8_t index)
+void pl_shader_reset_ex(struct pl_shader *sh, uint8_t index, uint8_t ident)
 {
     struct pl_shader new = {
         .ctx = sh->ctx,
@@ -82,6 +82,11 @@ void pl_shader_reset(struct pl_shader *sh, uint8_t ident, uint8_t index)
     *sh = new;
 }
 
+void pl_shader_reset(struct pl_shader *sh, uint8_t index)
+{
+    pl_shader_reset_ex(sh, index, 0);
+}
+
 bool pl_shader_is_failed(const struct pl_shader *sh)
 {
     return sh->failed;
@@ -179,6 +184,21 @@ ident_t sh_var(struct pl_shader *sh, struct pl_shader_var sv)
 
 ident_t sh_desc(struct pl_shader *sh, struct pl_shader_desc sd)
 {
+    // Skip re-attaching the same buffer desc twice
+    // FIXME: define aliases if the variable names differ
+    switch (sd.desc.type) {
+    case PL_DESC_BUF_UNIFORM:
+    case PL_DESC_BUF_STORAGE:
+    case PL_DESC_BUF_TEXEL_UNIFORM:
+    case PL_DESC_BUF_TEXEL_STORAGE:
+        for (int i = 0; i < sh->res.num_descriptors; i++) {
+            if (sh->res.descriptors[i].object == sd.object)
+                return (ident_t) sh->res.descriptors[i].desc.name;
+        }
+
+    default: break;
+    }
+
     sd.desc.name = sh_fresh(sh, sd.desc.name);
     TARRAY_APPEND(sh, sh->res.descriptors, sh->res.num_descriptors, sd);
     return (ident_t) sd.desc.name;
@@ -407,6 +427,12 @@ static ident_t sh_split(struct pl_shader *sh)
         sh->buffers[SH_BUF_BODY].start[0] = '\0'; // for sanity / efficiency
     }
 
+    if (sh->buffers[SH_BUF_FOOTER].len) {
+        bstr_xappend(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_FOOTER]);
+        sh->buffers[SH_BUF_FOOTER].len = 0;
+        sh->buffers[SH_BUF_FOOTER].start[0] = '\0';
+    }
+
     GLSLH("%s }\n", retvals[sh->res.output]);
     return name;
 }
@@ -629,7 +655,7 @@ ident_t sh_lut(struct pl_shader *sh, struct pl_shader_obj **obj,
     int dims = depth ? 3 : height ? 2 : 1;
 
     int texdim = 0;
-    int max_tex_dim[] = {
+    uint32_t max_tex_dim[] = {
         gpu ? gpu->limits.max_tex_1d_dim : 0,
         gpu ? gpu->limits.max_tex_2d_dim : 0,
         gpu ? gpu->limits.max_tex_3d_dim : 0,
