@@ -415,27 +415,38 @@ void pl_shader_unsigmoidize(struct pl_shader *sh,
          1.0 / scale, slope, center, offset / scale);
 }
 
-// Applies the OOTF / inverse OOTF
-static void pl_shader_ootf(struct pl_shader *sh, enum pl_color_light light,
-                           ident_t luma, float peak)
+static ident_t sh_luma_coeffs(struct pl_shader *sh, enum pl_color_primaries prim)
 {
-    if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
-        return;
+    struct pl_matrix3x3 rgb2xyz;
+    rgb2xyz = pl_get_rgb2xyz_matrix(pl_raw_primaries_get(prim));
+    return sh_var(sh, (struct pl_shader_var) {
+        .var = pl_var_vec3("luma_coeffs"),
+        .data = rgb2xyz.m[1], // RGB->Y vector
+    });
+}
 
-    if (!light || light == PL_COLOR_LIGHT_DISPLAY)
+// Applies the OOTF / inverse OOTF - including the sig_scale adaptation
+static void pl_shader_ootf(struct pl_shader *sh, struct pl_color_space csp)
+{
+    if (csp.sig_scale != 1.0)
+        GLSL("color.rgb *= vec3(%f); \n", csp.sig_scale);
+
+    if (!csp.light || csp.light == PL_COLOR_LIGHT_DISPLAY)
         return;
 
     GLSL("// pl_shader_ootf                \n"
          "color.rgb = max(color.rgb, 0.0); \n");
 
-    switch (light)
+    switch (csp.light)
     {
     case PL_COLOR_LIGHT_SCENE_HLG: {
         // HLG OOTF from BT.2100, tuned to the indicated peak
-        float gamma = 1.2 + 0.42 * log10(peak * PL_COLOR_REF_WHITE / 1000.0);
+        float gamma = 1.2 + 0.42 * log10(csp.sig_peak * PL_COLOR_REF_WHITE / 1000.0);
         gamma = PL_MAX(gamma, 1.0);
         GLSL("color.rgb *= vec3(%f * pow(dot(%s, color.rgb), %f));\n",
-             peak / pow(12, gamma), luma, gamma - 1.0);
+             csp.sig_peak / pow(12, gamma),
+             sh_luma_coeffs(sh, csp.primaries),
+             gamma - 1.0);
         break;
     }
     case PL_COLOR_LIGHT_SCENE_709_1886:
@@ -457,25 +468,25 @@ static void pl_shader_ootf(struct pl_shader *sh, enum pl_color_light light,
     }
 }
 
-static void pl_shader_inverse_ootf(struct pl_shader *sh,
-                                   enum pl_color_light light, ident_t luma,
-                                   float peak)
+static void pl_shader_inverse_ootf(struct pl_shader *sh, struct pl_color_space csp)
 {
-    if (!light || light == PL_COLOR_LIGHT_DISPLAY)
-        return;
+    if (!csp.light || csp.light == PL_COLOR_LIGHT_DISPLAY)
+        goto skip;
 
     GLSL("// pl_shader_inverse_ootf        \n"
          "color.rgb = max(color.rgb, 0.0); \n");
 
-    switch (light)
+    switch (csp.light)
     {
     case PL_COLOR_LIGHT_SCENE_HLG: {
-        float gamma = 1.2 + 0.42 * log10(peak * PL_COLOR_REF_WHITE / 1000.0);
+        float gamma = 1.2 + 0.42 * log10(csp.sig_peak * PL_COLOR_REF_WHITE / 1000.0);
         gamma = PL_MAX(gamma, 1.0);
         GLSL("color.rgb *= vec3(1.0/%f);                                \n"
              "color.rgb /= vec3(max(1e-6, pow(dot(%s, color.rgb),       \n"
              "                                %f)));                    \n",
-             peak / pow(12, gamma), luma, (gamma - 1.0) / gamma);
+             csp.sig_peak / pow(12, gamma),
+             sh_luma_coeffs(sh, csp.primaries),
+             (gamma - 1.0) / gamma);
         break;
     }
     case PL_COLOR_LIGHT_SCENE_709_1886:
@@ -492,19 +503,22 @@ static void pl_shader_inverse_ootf(struct pl_shader *sh,
     default:
         abort();
     }
+
+skip:
+    if (csp.sig_scale != 1.0)
+        GLSL("color.rgb *= vec3(1.0 / %f); \n", csp.sig_scale);
 }
 
-const struct pl_color_map_params pl_color_map_default_params = {
-    .intent                  = PL_INTENT_RELATIVE_COLORIMETRIC,
-    .tone_mapping_algo       = PL_TONE_MAPPING_HABLE,
-    .tone_mapping_desaturate = 0.5,
-    .peak_detect_frames      = 63,
-    .scene_threshold         = 0.2,
+const struct pl_peak_detect_params pl_peak_detect_default_params = {
+    .smoothing_period       = 100.0,
+    .scene_threshold_low    = 5.5,
+    .scene_threshold_high   = 10.0,
 };
 
 struct sh_peak_obj {
     const struct pl_gpu *gpu;
     const struct pl_buf *buf;
+    struct pl_shader_desc desc;
 };
 
 static void sh_peak_uninit(const struct pl_gpu *gpu, void *ptr)
@@ -514,232 +528,214 @@ static void sh_peak_uninit(const struct pl_gpu *gpu, void *ptr)
     *obj = (struct sh_peak_obj) {0};
 }
 
-static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
-                            const struct pl_color_map_params *params)
+static inline float iir_coeff(float rate)
 {
-    int frames = params->peak_detect_frames;
-    if (!state || !frames)
-        return;
+    float a = 1.0 - cos(1.0 / rate);
+    return sqrt(a*a + 2*a) - a;
+}
 
-    if (frames < 0 || frames > 1000) {
-        SH_FAIL(sh, "Parameter peak_detect_frames must be >= 0 and <= 1000 "
-                "(was %d).", frames);
-        return;
+bool pl_shader_detect_peak(struct pl_shader *sh,
+                           struct pl_color_space csp,
+                           struct pl_shader_obj **state,
+                           const struct pl_peak_detect_params *params)
+{
+    params = PL_DEF(params, &pl_peak_detect_default_params);
+    if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
+        return false;
+
+    if (!sh_try_compute(sh, 8, 8, true, 2 * sizeof(int32_t))) {
+        PL_ERR(sh, "HDR peak detection requires compute shaders!");
+        return false;
     }
 
     struct sh_peak_obj *obj;
     obj = SH_OBJ(sh, state, PL_SHADER_OBJ_PEAK_DETECT, struct sh_peak_obj,
                  sh_peak_uninit);
     if (!obj)
-        return;
-
-    if (!sh_try_compute(sh, 8, 8, true, sizeof(uint32_t))) {
-        PL_WARN(sh, "HDR peak detection requires compute shaders.. disabling");
-        return;
-    }
+        return false;
 
     const struct pl_gpu *gpu = sh->gpu;
     obj->gpu = gpu;
 
-    struct pl_var idx, num, ctr, max, avg;
-    idx = pl_var_uint(sh_fresh(sh, "index"));
-    num = pl_var_uint(sh_fresh(sh, "number"));
-    ctr = pl_var_uint(sh_fresh(sh, "counter"));
-    max = pl_var_uint(sh_fresh(sh, "frames_max"));
-    avg = pl_var_uint(sh_fresh(sh, "frames_avg"));
-    max.dim_a = avg.dim_a = frames + 1;
+    if (!obj->buf) {
+        obj->desc = (struct pl_shader_desc) {
+            .desc = {
+                .name   = "PeakDetect",
+                .type   = PL_DESC_BUF_STORAGE,
+            },
+        };
 
-    struct pl_var max_total, avg_total;
-    max_total = pl_var_uint(sh_fresh(sh, "max_total"));
-    avg_total = pl_var_uint(sh_fresh(sh, "avg_total"));
+        bool ok = true;
+        ok &= sh_buf_desc_append(obj, gpu, &obj->desc, NULL, pl_var_vec2("average"));
+        ok &= sh_buf_desc_append(obj, gpu, &obj->desc, NULL, pl_var_int("frame_sum"));
+        ok &= sh_buf_desc_append(obj, gpu, &obj->desc, NULL, pl_var_int("frame_max"));
+        ok &= sh_buf_desc_append(obj, gpu, &obj->desc, NULL, pl_var_uint("counter"));
 
-    // Attempt packing the peak detection SSBO
-    struct pl_shader_desc ssbo = {
-        .desc = {
-            .name   = "PeakDetect",
-            .type   = PL_DESC_BUF_STORAGE,
-            .access = PL_DESC_ACCESS_READWRITE,
-        },
-    };
-
-    bool ok = true;
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, idx);
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, num);
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, ctr);
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, max);
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, avg);
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, max_total);
-    ok &= sh_buf_desc_append(sh->tmp, gpu, &ssbo, NULL, avg_total);
-
-    if (!ok) {
-        PL_WARN(sh, "HDR peak detection exhausts device limits.. disabling");
-        talloc_free(ssbo.buffer_vars);
-        return;
-    }
-
-    // Create the SSBO if necessary
-    size_t size = sh_buf_desc_size(&ssbo);
-    if (!obj->buf || obj->buf->params.size != size) {
-        PL_TRACE(sh, "(Re)creating HDR peak detection SSBO");
+        if (!ok) {
+            PL_ERR(sh, "HDR peak detection exhausts device limits!");
+            return false;
+        }
 
+        // Create the SSBO
+        size_t size = sh_buf_desc_size(&obj->desc);
         void *data = talloc_zero_size(NULL, size);
-        pl_buf_destroy(gpu, &obj->buf);
         obj->buf = pl_buf_create(gpu, &(struct pl_buf_params) {
             .type = PL_BUF_STORAGE,
             .size = size,
             .initial_data = data,
         });
+        obj->desc.object = obj->buf;
         talloc_free(data);
     }
 
     if (!obj->buf) {
-        SH_FAIL(sh, "Failed creating peak detection SSBO!");
-        return;
+        PL_ERR(sh, "Failed creating peak detection SSBO!");
+        return false;
     }
 
     // Attach the SSBO and perform the peak detection logic
-    ssbo.object = obj->buf;
-    sh_desc(sh, ssbo);
+    obj->desc.desc.access = PL_DESC_ACCESS_READWRITE;
+    sh_desc(sh, obj->desc);
+    GLSL("// pl_shader_detect_peak \n"
+         "{                        \n"
+         "vec4 color_orig = color; \n");
+
+    // Decode the color into linear light absolute scale representation
+    pl_color_space_infer(&csp);
+    pl_shader_linearize(sh, csp.transfer);
+    pl_shader_ootf(sh, csp);
 
     // For performance, we want to do as few atomic operations on global
     // memory as possible, so use an atomic in shmem for the work group.
-    ident_t wg_sum = sh_fresh(sh, "wg_sum");
-    GLSLH("shared uint %s;\n", wg_sum);
-    GLSL("%s = 0;\n", wg_sum);
+    ident_t wg_sum = sh_fresh(sh, "wg_sum"), wg_max = sh_fresh(sh, "wg_max");
+    GLSLH("shared int %s;   \n", wg_sum);
+    GLSLH("shared int %s;   \n", wg_max);
+    GLSL("%s = 0; %s = 0;   \n"
+         "barrier();        \n",
+         wg_sum, wg_max);
+
+    // Chosen to avoid overflowing on an 8K buffer
+    const float log_min = 1e-3, log_scale = 400.0, sig_scale = 10000.0;
 
     // Have each thread update the work group sum with the local value
-    GLSL("barrier();                     \n"
-         "atomicAdd(%s, uint(sig * %f)); \n",
-         wg_sum, PL_COLOR_REF_WHITE);
+    GLSL("float sig_max = max(max(color.r, color.g), color.b);  \n"
+         "float sig_log = log(max(sig_max, %f));                \n"
+         "atomicAdd(%s, int(sig_log * %f));                     \n"
+         "atomicMax(%s, int(sig_max * %f));                     \n"
+         "memoryBarrierShared();                                \n"
+         "barrier();                                            \n"
+         "color = color_orig;                                   \n"
+         "}                                                     \n",
+         log_min,
+         wg_sum, log_scale,
+         wg_max, sig_scale);
 
-    // Have one thread per work group update the global atomics. We use the
-    // work group average even for the global max, to make the values slightly
-    // more stable and smooth out tiny super-highlights.
-    GLSL("memoryBarrierShared();                                            \n"
-         "barrier();                                                        \n"
-         "if (gl_LocalInvocationIndex == 0) {                               \n"
-         "    uint wg_avg = %s / (gl_WorkGroupSize.x * gl_WorkGroupSize.y); \n"
-         "    atomicMax(%s[%s], wg_avg);                                    \n"
-         "    atomicAdd(%s[%s], wg_avg);                                    \n"
-         "}                                                                 \n",
-         wg_sum,
-         max.name, idx.name,
-         avg.name, idx.name);
+    // Have one thread per work group update the global atomics. Do this
+    // at the end of the shader to avoid clobbering `average`, in case the
+    // state object will be used by the same pass.
+    GLSLF("if (gl_LocalInvocationIndex == 0) {                                  \n"
+          "    int wg_avg = %s / int(gl_WorkGroupSize.x * gl_WorkGroupSize.y);  \n"
+          "    atomicAdd(frame_sum, wg_avg);                                    \n"
+          "    atomicMax(frame_max, %s);                                        \n"
+          "    memoryBarrierBuffer();                                           \n"
+          "    barrier();                                                       \n",
+          wg_sum, wg_max);
 
-    // Update the sig_peak/sig_avg from the old SSBO state
-    GLSL("if (%s > 0) {                                  \n"
-         "    float peak = float(%s) / (%f * float(%s)); \n"
-         "    float avg  = float(%s) / (%f * float(%s)); \n"
-         "    sig_peak   = max(1.0, peak);               \n"
-         "    sig_avg    = max(0.25, avg);               \n"
-         "}                                              \n",
-         num.name,
-         max_total.name, PL_COLOR_REF_WHITE, num.name,
-         avg_total.name, PL_COLOR_REF_WHITE, num.name);
+    // Finally, to update the global state per dispatch, we increment a counter
+    GLSLF("    uint num_wg = gl_NumWorkGroups.x * gl_NumWorkGroups.y;           \n"
+          "    if (atomicAdd(counter, 1) == num_wg - 1) {                       \n"
+          "        vec2 cur = vec2(float(frame_sum) / float(num_wg), frame_max);\n"
+          "        cur *= vec2(1.0 / %f, 1.0 / %f);                             \n"
+          "        cur.x = exp(cur.x);                                          \n",
+          log_scale, sig_scale);
 
-    // Finally, to update the global state, we increment a counter per dispatch
-    GLSL("memoryBarrierBuffer();                                                \n"
-         "barrier();                                                            \n"
-         "uint num_wg = gl_NumWorkGroups.x * gl_NumWorkGroups.y;                \n"
-         "if (gl_LocalInvocationIndex == 0 && atomicAdd(%s, 1) == num_wg - 1) { \n"
-         "    %s = 0;                                                           \n"
-         // Divide out the workgroup sum by the number of work groups
-         "    %s[%s] /= num_wg;                                                 \n",
-         ctr.name, ctr.name,
-         avg.name, idx.name);
+    // Set the initial value accordingly if it contains no data
+    GLSLF("        if (average.y == 0.0) \n"
+          "            average = cur;    \n");
 
-    // Scene change detection
-    if (params->scene_threshold > 0) {
-    GLSL("    uint cur_max = %s[%s];                  \n"
-         "    uint cur_avg = %s[%s];                  \n"
-         "    int diff = int(%s * cur_avg) - int(%s); \n"
-         "    if (abs(diff) > %s * %d) {              \n"
-         "        %s = 0;                             \n"
-         "        %s = %s = 0;                        \n"
-         "        for (uint i = 0; i < %d; i++)       \n"
-         "            %s[i] = %s[i] = 0;              \n"
-         "       %s[%s] = cur_max;                    \n"
-         "       %s[%s] = cur_avg;                    \n"
-         "    }                                       \n",
-         max.name, idx.name,
-         avg.name, idx.name,
-         num.name, avg_total.name,
-         num.name, (int) (params->scene_threshold * PL_COLOR_REF_WHITE),
-         num.name,
-         max_total.name, avg_total.name,
-         frames + 1,
-         max.name, avg.name,
-         max.name, idx.name,
-         avg.name, idx.name);
+    // Use an IIR low-pass filter to smooth out the detected values
+    GLSLF("        average += %f * (cur - average); \n",
+          iir_coeff(PL_DEF(params->smoothing_period, 100.0)));
+
+    // Scene change hysteresis
+    float log_db = 10.0 / log(10.0);
+    if (params->scene_threshold_low > 0 && params->scene_threshold_high > 0) {
+        GLSLF("    float delta = abs(log(cur.x / average.x));               \n"
+              "    average = mix(average, cur, smoothstep(%f, %f, delta));  \n",
+              params->scene_threshold_low / log_db,
+              params->scene_threshold_high / log_db);
     }
 
-    // Add the current frame, then subtract and reset the next frame
-    GLSL("    uint next = (%s + 1) %% %d; \n"
-         "    %s += %s[%s] - %s[next];    \n"
-         "    %s += %s[%s] - %s[next];    \n"
-         "    %s[next] = %s[next] = 0;    \n",
-         idx.name, frames + 1,
-         max_total.name, max.name, idx.name, max.name,
-         avg_total.name, avg.name, idx.name, avg.name,
-         max.name, avg.name);
+    // Reset SSBO state for the next frame
+    GLSLF("        frame_sum = 0;            \n"
+          "        frame_max = 0;            \n"
+          "        counter = 0;              \n"
+          "        memoryBarrierBuffer();    \n"
+          "    }                             \n"
+          "}                                 \n");
 
-    // Update the index and count
-    GLSL("    %s = next;             \n"
-         "    %s = min(%s + 1, %d);  \n"
-         "    memoryBarrierBuffer(); \n"
-         "}                          \n",
-         idx.name,
-         num.name, num.name, frames);
+    return true;
 }
 
+const struct pl_color_map_params pl_color_map_default_params = {
+    .intent                 = PL_INTENT_RELATIVE_COLORIMETRIC,
+    .tone_mapping_algo      = PL_TONE_MAPPING_HABLE,
+    .desaturation_strength  = 0.75,
+    .desaturation_exponent  = 1.50,
+    .desaturation_base      = 0.18,
+};
+
 static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
-                               struct pl_color_space dst, ident_t luma,
+                               struct pl_color_space dst,
                                struct pl_shader_obj **peak_detect_state,
                                const struct pl_color_map_params *params)
 {
-    GLSL("// pl_shader_tone_map\n");
+    GLSL("// pl_shader_tone_map \n");
 
     // To prevent discoloration due to out-of-bounds clipping, we need to make
     // sure to reduce the value range as far as necessary to keep the entire
     // signal in range, so tone map based on the brightest component.
-    GLSL("float sig = max(max(color.r, color.g), color.b); \n"
-         "float sig_peak = %f;                             \n"
-         "float sig_avg = %f;                              \n",
-         src.sig_peak, src.sig_avg);
+    GLSL("int sig_idx = 0;                              \n"
+         "if (color[1] > color[sig_idx]) sig_idx = 1;   \n"
+         "if (color[2] > color[sig_idx]) sig_idx = 2;   \n"
+         "float sig_max = color[sig_idx];               \n"
+         "float sig_peak = %f;                          \n"
+         "float sig_avg = %f;                           \n",
+         src.sig_peak * src.sig_scale,
+         src.sig_avg * src.sig_scale);
 
-    // HDR peak detection is done before scaling based on the dst.sig_peak/avg
-    // in order to make the detected values stable / averageable.
-    hdr_update_peak(sh, peak_detect_state, params);
+    // Update the variables based on values from the peak detection buffer
+    if (peak_detect_state) {
+        struct sh_peak_obj *obj;
+        obj = SH_OBJ(sh, peak_detect_state, PL_SHADER_OBJ_PEAK_DETECT,
+                     struct sh_peak_obj, sh_peak_uninit);
+        if (obj && obj->buf) {
+            obj->desc.desc.access = PL_DESC_ACCESS_READONLY;
+            sh_desc(sh, obj->desc);
+            GLSL("sig_avg  = average.x; \n"
+                 "sig_peak = average.y; \n");
+        }
+    }
 
-    // Rescale the variables in order to bring it into a representation where
+    // Rescale the input in order to bring it into a representation where
     // 1.0 represents the dst_peak. This is because all of the tone mapping
     // algorithms are defined in such a way that they map to the range [0.0, 1.0].
-    if (dst.sig_peak > 1.0) {
-        GLSL("sig *= 1.0/%f;      \n"
-             "sig_peak *= 1.0/%f; \n",
-             dst.sig_peak, dst.sig_peak);
+    float dst_range = dst.sig_peak * dst.sig_scale;
+    if (dst_range > 1.0) {
+        GLSL("color.rgb *= 1.0 / %f; \n"
+             "sig_peak *= 1.0 / %f;  \n",
+             dst_range, dst_range);
     }
 
-    // Store the original signal level for later re-use
-    GLSL("float sig_orig = sig;\n");
+    // Rename `color.rgb` to something shorter for conciseness
+    GLSL("vec3 sig = color.rgb; \n"
+         "vec3 sig_orig = sig;  \n");
 
     // Scale the signal to compensate for differences in the average brightness
-    GLSL("float slope = min(1.0, %f / sig_avg); \n"
-         "sig *= slope;                         \n"
-         "sig_peak *= slope;                    \n",
-         dst.sig_avg);
-
-    // Desaturate the color using a coefficient dependent on the signal level.
-    // Do this after the brightness adjustment stage to prevent catastrophic
-    // saturation loss on overly bright HDR sources.
-    if (params->tone_mapping_desaturate > 0) {
-        GLSL("float luma = %f * dot(%s, color.rgb);                 \n"
-             "float coeff = max(sig - 0.18, 1e-6) / max(sig, 1e-6); \n"
-             "coeff = pow(coeff, %f);                               \n"
-             "color.rgb = mix(color.rgb, vec3(luma), coeff);        \n"
-             "sig = mix(sig, luma * slope, coeff);                  \n",
-             1.0 / dst.sig_peak, luma, 10.0 / params->tone_mapping_desaturate);
-    }
+    GLSL("float slope = min(%f, %f / sig_avg); \n"
+         "sig *= slope;                        \n"
+         "sig_peak *= slope;                   \n",
+         PL_DEF(params->max_boost, 1.0), dst.sig_avg * dst.sig_scale);
 
     float param = params->tone_mapping_param;
     switch (params->tone_mapping_algo) {
@@ -759,15 +755,17 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
              "    float b = (j*j - 2.0*j*sig_peak + sig_peak) /                 \n"
              "              max(1e-6, sig_peak - 1.0);                          \n"
              "    float scale = (b*b + 2.0*b*j + j*j) / (b-a);                  \n"
-             "    sig = sig > j ? (scale * (sig + a) / (sig + b)) : sig;        \n"
+             "    sig = mix(sig, scale * (sig + vec3(a)) / (sig + vec3(b)),     \n"
+             "              %s(greaterThan(sig, vec3(j))));                     \n"
              "}                                                                 \n",
-             PL_DEF(param, 0.3));
+             PL_DEF(param, 0.3),
+             sh_bvec(sh, 3));
         break;
 
     case PL_TONE_MAPPING_REINHARD: {
         float contrast = PL_DEF(param, 0.5),
               offset = (1.0 - contrast) / contrast;
-        GLSL("sig = sig / (sig + %f);                   \n"
+        GLSL("sig = sig / (sig + vec3(%f));             \n"
              "float scale = (sig_peak + %f) / sig_peak; \n"
              "sig *= scale;                             \n",
              offset, offset);
@@ -777,19 +775,24 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
     case PL_TONE_MAPPING_HABLE: {
         float A = 0.15, B = 0.50, C = 0.10, D = 0.20, E = 0.02, F = 0.30;
         ident_t hable = sh_fresh(sh, "hable");
-        GLSLH("float %s(float x) {                                        \n"
-              "return ((x * (%f*x + %f)+%f)/(x * (%f*x + %f) + %f)) - %f; \n"
-              "}                                                          \n",
+        GLSLH("vec3 %s(vec3 x) {                                \n"
+              "    return (x * (%f*x + vec3(%f)) + vec3(%f)) /  \n"
+              "           (x * (%f*x + vec3(%f)) + vec3(%f))    \n"
+              "           - vec3(%f);                           \n"
+              "}                                                \n",
               hable, A, C*B, D*E, A, B, D*F, E/F);
-        GLSL("sig = %s(sig) / %s(sig_peak);\n", hable, hable);
+        GLSL("sig = %s(sig) / %s(vec3(sig_peak)).x;\n", hable, hable);
         break;
     }
 
     case PL_TONE_MAPPING_GAMMA:
-        GLSL("const float cutoff = 0.05, gamma = 1.0/%f;                     \n"
-             "float scale = pow(cutoff / sig_peak, gamma) / cutoff;          \n"
-             "sig = sig > cutoff ? pow(sig / sig_peak, gamma) : scale * sig; \n",
-             PL_DEF(param, 1.8));
+        GLSL("const float cutoff = 0.05, gamma = 1.0/%f;            \n"
+             "float scale = pow(cutoff / sig_peak, gamma) / cutoff; \n"
+             "sig = mix(scale * sig,                                \n"
+             "          pow(sig / sig_peak, vec3(gamma)),           \n"
+             "          %s(greaterThan(sig, vec3(cutoff))));        \n",
+             PL_DEF(param, 1.8),
+             sh_bvec(sh, 3));
         break;
 
     case PL_TONE_MAPPING_LINEAR:
@@ -800,10 +803,26 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
         abort();
     }
 
-    // Clip the final signal to the output range and apply the difference
-    // linearly to the RGB channels. (this prevents discoloration)
-    GLSL("sig = min(sig, 1.01);       \n"
-        "color.rgb *= sig / sig_orig; \n");
+    GLSL("sig = min(sig, 1.01);                                         \n"
+         "vec3 sig_lin = sig_orig * (sig[sig_idx] / sig_orig[sig_idx]); \n");
+
+    // Mix between the per-channel tone mapped `sig` and the linear tone
+    // mapped `sig_lin` based on the desaturation strength
+    if (params->desaturation_strength > 0.0) {
+        GLSL("float coeff = max(sig[sig_idx] - %f, 1e-6) /  \n"
+             "              max(sig[sig_idx], 1.0);         \n"
+             "coeff = %f * pow(coeff, %f);                  \n"
+             "color.rgb = mix(sig_lin, sig, coeff);         \n",
+             params->desaturation_base,
+             params->desaturation_strength,
+             params->desaturation_exponent);
+    } else {
+        GLSL("color.rgb = sig_lin; \n");
+    }
+
+    // Undo the normalization by `dst_peak`
+    if (dst_range > 1.0)
+        GLSL("color.rgb *= %f; \n", dst_range);
 }
 
 void pl_shader_color_map(struct pl_shader *sh,
@@ -836,15 +855,17 @@ void pl_shader_color_map(struct pl_shader *sh,
     if (!dst.transfer) {
         dst.transfer = src.transfer;
         if (pl_color_transfer_is_hdr(dst.transfer) ||
-                dst.transfer == PL_COLOR_TRC_LINEAR)
+            dst.transfer == PL_COLOR_TRC_LINEAR)
+        {
             dst.transfer = PL_COLOR_TRC_GAMMA22;
+        }
     }
 
     // Defaults the dest average based on the source average, unless the source
     // is HDR and the destination is not
     if (!dst.sig_avg) {
-        bool src_hdr = pl_color_transfer_is_hdr(src.transfer);
-        bool dst_hdr = pl_color_transfer_is_hdr(dst.transfer);
+        bool src_hdr = pl_color_space_is_hdr(src);
+        bool dst_hdr = pl_color_space_is_hdr(dst);
         if (!(src_hdr && !dst_hdr))
             dst.sig_avg = src.sig_avg;
     }
@@ -852,37 +873,15 @@ void pl_shader_color_map(struct pl_shader *sh,
     // Infer the remaining fields after making the above choices
     pl_color_space_infer(&dst);
 
-    // HLG's OOTF is parametrized by the sig peak, so enable OOTF conversion
-    // if necessary, even if the output light is the same
-    bool need_ootf = src.light != dst.light;
-    if (src.transfer == PL_COLOR_TRC_HLG && src.sig_peak != dst.sig_peak)
-        need_ootf = true;
-
     // All operations from here on require linear light as a starting point,
-    // so we linearize even if src.gamma == dst.gamma when one of the other
+    // so we linearize even if src.transfer == dst.transfer when one of the other
     // operations needs it
     bool need_linear = src.transfer != dst.transfer ||
                        src.primaries != dst.primaries ||
                        src.sig_peak > dst.sig_peak ||
                        src.sig_avg != dst.sig_avg ||
-                       need_ootf;
-
-    // Various operations need access to the src_luma and dst_luma respectively,
-    // so just always make them available if we're doing anything at all
-    ident_t src_luma = NULL, dst_luma = NULL;
-    if (need_linear) {
-        struct pl_matrix3x3 rgb2xyz;
-        rgb2xyz = pl_get_rgb2xyz_matrix(pl_raw_primaries_get(src.primaries));
-        src_luma = sh_var(sh, (struct pl_shader_var) {
-            .var  = pl_var_vec3("src_luma"),
-            .data = rgb2xyz.m[1], // RGB->Y vector
-        });
-        rgb2xyz = pl_get_rgb2xyz_matrix(pl_raw_primaries_get(dst.primaries));
-        dst_luma = sh_var(sh, (struct pl_shader_var) {
-            .var  = pl_var_vec3("dst_luma"),
-            .data = rgb2xyz.m[1], // RGB->Y vector
-        });
-    }
+                       src.sig_scale != dst.sig_scale ||
+                       src.light != dst.light;
 
     bool is_linear = prelinearized;
     if (need_linear && !is_linear) {
@@ -890,8 +889,12 @@ void pl_shader_color_map(struct pl_shader *sh,
         is_linear = true;
     }
 
-    if (need_ootf)
-        pl_shader_ootf(sh, src.light, src_luma, src.sig_peak);
+    if (need_linear)
+        pl_shader_ootf(sh, src);
+
+    // Tone map to rescale the signal average/peak if needed
+    if (src.sig_peak * src.sig_scale > dst.sig_peak * dst.sig_scale + 1e-6)
+        pl_shader_tone_map(sh, src, dst, peak_detect_state, params);
 
     // Adapt to the right colorspace (primaries) if necessary
     if (src.primaries != dst.primaries) {
@@ -906,23 +909,16 @@ void pl_shader_color_map(struct pl_shader *sh,
         }));
     }
 
-    // Tone map to rescale the signal average/peak if needed
-    if (src.sig_peak > dst.sig_peak + 1e-6)
-        pl_shader_tone_map(sh, src, dst, dst_luma, peak_detect_state, params);
-
     // Warn for remaining out-of-gamut colors if enabled
     if (params->gamut_warning) {
         GLSL("if (any(greaterThan(color.rgb, vec3(%f + 0.005))) ||\n"
              "    any(lessThan(color.rgb, vec3(-0.005))))\n"
-             "    color.rgb = vec3(1.0) - color.rgb; // invert\n",
-             dst.sig_peak);
+             "    color.rgb = vec3(%f) - color.rgb; // invert\n",
+             dst.sig_peak * dst.sig_scale, src.sig_peak * src.sig_scale);
     }
 
-    if (need_ootf)
-        pl_shader_inverse_ootf(sh, dst.light, dst_luma, dst.sig_peak);
-
-    if (params->hdr_simulation && !pl_color_transfer_is_hdr(dst.transfer))
-        GLSL("color.rgb *= vec3(%f);\n", 1.0 / dst.sig_peak);
+    if (need_linear)
+        pl_shader_inverse_ootf(sh, dst);
 
     if (is_linear)
         pl_shader_delinearize(sh, dst.transfer);
